from .ca_implementation import BLUvoCa


class BLUvoCaHyundai(BLUvoCa):

    def __init__(self, bluvo_config):
        super().__init__(bluvo_config)
        self.bluvo_config.base_url = 'https://mybluelink.ca/tods/api/'



import requests
import json
import logging

from datetime import datetime

from .bluvo_interface import BLUvoInternalInterface, BLUvoSession
from .helpers.either import Either
from .vehicles.ca_vehicle import BLUvoCaVehicle


class BLUvoCa(BLUvoInternalInterface):

    def __init__(self, bluvo_config):
        super().__init__(bluvo_config)
        self.bluvo_config = bluvo_config
        self.session = BLUvoSession(None, None, 0)

    def request(self, cmd, data=None, headers=None) -> Either[dict, str]:

        if headers is None:
            headers = {}

        if data is None:
            data = {}

        default_headers = {
            'content-type': 'application/json;charset=UTF-8',
            'accept': 'application/json, text/plain, */*',
            'accept-encoding': 'gzip, deflate, br',
            "from": 'SPA',
            "language": "1",
            "offset": "-5",
            "accessToken": self.session.access_token
        }

        default_data = {
            'pin': self.bluvo_config.pin
        }

        req = requests.post(self.bluvo_config.base_url + cmd,
                            headers={**default_headers, **headers},
                            data=json.dumps({**default_data, **data}))

        print("{}: {}".format(cmd, req.status_code))

        if req.status_code != requests.codes.ok:
            logging.error('%s: %s' % ("HttpError", req.status_code))
            return Either[dict, str].right("HttpError")

        response = req.json()

        if "error" in response:
            logging.error('%s: %s' % ("InvalidRequest", response["error"]))
            return Either[dict, str].right("InvalidRequest")

        print(req.json())
        return Either[dict, str].left(req.json())

    def handle_session_login(self, res):
        self.session = BLUvoSession(
            res['result']['accessToken'],
            res['result']['refreshToken'],
            res['result']['expireIn'])

    def handle_exception(self):
        pass

    def login(self):

        response = self.request("lgn", data={
            'loginId': self.bluvo_config.username,
            'password': self.bluvo_config.password
        })

        response.match(self.handle_session_login, self.handle_exception)

    def refresh_token(self):
        now = datetime.now()
        timestamp = datetime.timestamp(now)
        should_refresh = timestamp - self.session.token_expires_at <= 10

        if self.session.refresh_token and should_refresh:
            response = self.request("vrfytnc")

            response.match(self.handle_session_login, self.handle_exception)

    def verify_pin(self):
        response = self.request("vrfypin")

        pin_token = response.match(
            lambda res: res['result']['pAuth'],
            self.handle_exception)

        self.session.set_pin_token(pin_token)

    def get_vehicles(self):
        response = self.request("vhcllst")
        # self.verify_pin()
        # location_response = self.request("fndmcr", {'pAuth': self.session.pin_token})

        return response.match(
            lambda res: [
                BLUvoCaVehicle(self, self.bluvo_config, vehicle_info)
                for vehicle_info in res['result']['vehicles']],
            self.handle_exception)


from typing import List

from .bluvo_interface import BLUvoInternalInterface, BLUvoConfig, VehicleInterface
from .helpers.either import Either


class BLUvoEu(BLUvoInternalInterface):

    def __init__(self, bluvo_config: BLUvoConfig):
        super().__init__(bluvo_config)

    def login(self):
        pass

    def verify_pin(self, pin: str):
        pass

    def get_vehicles(self) -> List[VehicleInterface]:
        pass

    def request(self, cmd, data=None, headers=None) -> Either[dict, str]:
        pass


from . import test_config as cfg

from .api.bluvo_interface import VehicleStatus, VehicleLocation, BLUvoConfig
from .api.helpers.constants import Regions, Makes
from . import BLUvo


def print_status(status: VehicleStatus):
    print (status.chassis.locked)
    print (status.climate.temperature_set_point)
    pass


def print_location(location: VehicleLocation):
    print("{}, {}".format(location.latitude, location.longitude))


def handle_exception(ex):
    print(ex)


if __name__ == "__main__":
    config = BLUvoConfig(Regions.CA, Makes.KIA, cfg.username, cfg.password, cfg.pin)
    ctrl = BLUvo(config)

    ctrl.login()
    vehicles = ctrl.get_vehicles()

    first_vehicle = vehicles[0]

    first_vehicle.location().match(print_location, handle_exception)
    first_vehicle.status().match(print_status, handle_exception)



from typing import List

from .api.bluvo_interface import BLUvoInterface, BLUvoConfig, VehicleInterface
from .api.ca_hyundai_implementation import BLUvoCaHyundai
from .api.ca_kia_implementation import BLUvoCaKia
from .api.helpers.constants import Regions, Makes


class BLUvo(BLUvoInterface):

    impl: BLUvoInterface

    def __init__(self, bluvo_config: BLUvoConfig):
        super().__init__(bluvo_config)

        if bluvo_config.region == Regions.CA and bluvo_config.make == Makes.KIA:
            self.impl = BLUvoCaKia(bluvo_config)
        elif bluvo_config.region == Regions.CA and bluvo_config.make == Makes.HYUNDAI:
            self.impl = BLUvoCaHyundai(bluvo_config)
        # if bluvo_config.region == Regions.US and bluvo_config.make == Makes.KIA:
        #     self.impl = BLUvoUsKia(bluvo_config)
        # if bluvo_config.region == Regions.US and bluvo_config.make == Makes.HYUNDAI:
        #     self.impl = BLUvoUsHyundai(bluvo_config)
        # if bluvo_config.region == Regions.EU and bluvo_config.make == Makes.KIA:
        #     self.impl = BLUvoEuKia(bluvo_config)
        # if bluvo_config.region == Regions.EU and bluvo_config.make == Makes.HYUNDAI:
        #     self.impl = BLUvoEuHyundai(bluvo_config)
        else:
            raise Exception("No Region or Make Implementation")

    def verify_pin(self, pin: str):
        self.impl.verify_pin(pin)

    def login(self):
        self.impl.login()

    def get_vehicles(self) -> List[VehicleInterface]:
        return self.impl.get_vehicles()



from ..bluvo_interface import VehicleInterface, BLUvoConfig, VehicleStatus, Chassis, Quadrants, Climate, Engine, \
    VehicleLocation
from ..helpers.dict_helpers import dict_get_path
from ..helpers.either import Either
from ..helpers.temperature_helpers import temp_code_to_celsius


class BLUvoCaVehicle(VehicleInterface):

    def __init__(self, ctrl, bluvo_config: BLUvoConfig, vehicle_info):
        super().__init__(bluvo_config)

        self.ctrl = ctrl
        self.id = vehicle_info["vehicleId"]

    def status(self, refresh=False) -> Either[VehicleStatus, str]:
        response = self.ctrl.request("lstvhclsts" if refresh else "rltmvhclsts", headers={
            "vehicleId": self.id
        })
        print(response)
        return response.match(self.handle_status, lambda ex: Either[VehicleStatus, str].right(ex))

    def handle_status(self, res):
        status = res["result"]["status"]

        return Either[VehicleStatus, str].left(VehicleStatus(
            chassis=Chassis(
                hood_open=status["hoodOpen"],
                trunk_open=status["trunkOpen"],
                locked=status["doorLock"],
                open_doors=Quadrants(
                    front_left=status["doorOpen"]["frontLeft"],
                    front_right=status["doorOpen"]["frontRight"],
                    back_left=status["doorOpen"]["backLeft"],
                    back_right=status["doorOpen"]["backRight"]
                ),
                tire_pressure_warning_lamp=Quadrants(
                    front_left=dict_get_path(status, "tirePressureLamp", "tirePressureWarningLampFrontLeft", default=False),
                    front_right=dict_get_path(status, "tirePressureLamp", "tirePressureWarningLampFrontRight", default=False),
                    back_left=dict_get_path(status, "tirePressureLamp", "tirePressureWarningLampRearLeft", default=False),
                    back_right=dict_get_path(status, "tirePressureLamp", "tirePressureWarningLampRearRight", default=False)
                )
            ),
            climate=Climate(
                active=status["airCtrlOn"],
                steering_wheel_heat=status["steerWheelHeat"] if "steerWheelHeat" in status else False,
                side_mirror_heat=False,
                rear_window_heat=status["sideBackWindowHeat"] if "sideBackWindowHeat" in status else False,
                defrost=status["defrost"],
                temperature_set_point=temp_code_to_celsius(dict_get_path(status, "airTemp", "value", default=None)),
                temperature_unit=dict_get_path(status, "airTemp", "unit", default=0),
            ),
            engine=Engine(
                ignition=status["engine"],
                adaptive_cruise_control=status["acc"],
                range=dict_get_path(status, "dte", "value", default=0),
                charging=dict_get_path(status, "evStatus", "batteryCharge", default=False),
                battery_charge_12v=dict_get_path(status, "battery", "batSoc", default=False),
                battery_charge_hv=dict_get_path(status, "evStatus", "batteryStatus", default=False)
            )
        ))

    def location(self):
        self.ctrl.verify_pin()
        response = self.ctrl.request("fndmcr", headers={
            "vehicleId": self.id
        })
        print(response)
        return response.match(self.handle_location, lambda ex: Either[VehicleStatus, str].right(ex))

    def handle_location(self, res):
        coordinates = res["result"]["coord"]

        return Either[VehicleLocation, str].left(VehicleLocation(
            coordinates["lat"],
            coordinates["lon"],
            coordinates["alt"]
        ))






import decimal
from itertools import islice


def float_range(start, stop, step):
  while start <= stop:
    yield decimal.Decimal(start)
    start += decimal.Decimal(step)


def temp_code_to_celsius(code: str) -> decimal.Decimal:
    if code:
        hex_temp_index = code[0:2]
        temp_range = float_range(14, 30, 0.5)
        temp_index = int(hex_temp_index, 16)

        return next(islice(temp_range, temp_index, None))
    else:
        return decimal.Decimal(0.0)


def celsius_to_temp_code(temp: decimal.Decimal) -> str:
    temp_range = list(float_range(14, 30, 0.5))
    temp_index = temp_range.index(temp)
    return hex(temp_index).split('x')[-1].upper() + "H"



from abc import abstractmethod
from datetime import datetime
from decimal import Decimal
from typing import List

from .helpers.constants import Regions, Makes
from .helpers.either import Either


class Engine(object):

    def __init__(self,
                 ignition: bool,
                 battery_charge_12v: int,
                 battery_charge_hv: int,
                 charging: bool,
                 range: int,
                 adaptive_cruise_control: bool):
        self.ignition = ignition
        self.battery_charge_12v = battery_charge_12v
        self.battery_charge_hv = battery_charge_hv
        self.charging = charging
        self.range = range
        self.adaptive_cruise_control = adaptive_cruise_control


class Climate(object):
    def __init__(self,
                 active: bool,
                 steering_wheel_heat: bool,
                 side_mirror_heat: bool,
                 rear_window_heat: bool,
                 temperature_set_point: Decimal,
                 temperature_unit: int,
                 defrost: bool):
        self.active = active
        self.steering_wheel_heat = steering_wheel_heat
        self.side_mirror_heat = side_mirror_heat
        self.rear_window_heat = rear_window_heat
        self.temperature_set_point = temperature_set_point
        self.temperature_unit = temperature_unit
        self.defrost = defrost


class Quadrants(object):
    def __init__(self, front_left: bool, front_right: bool, back_left: bool, back_right: bool):
        self.front_left = front_left
        self.front_right = front_right
        self.back_left = back_left
        self.back_right = back_right


class Chassis(object):
    def __init__(self,
                 hood_open: bool,
                 trunk_open: bool,
                 locked: bool,
                 open_doors: Quadrants,
                 tire_pressure_warning_lamp: Quadrants):
        self.hood_open = hood_open
        self.trunk_open = trunk_open
        self.locked = locked
        self.open_doors = open_doors
        self.tire_pressure_warning_lamp = tire_pressure_warning_lamp


class VehicleStatus(object):
    def __init__(self, engine: Engine, climate: Climate, chassis: Chassis):
        self.engine = engine
        self.climate = climate
        self.chassis = chassis


class VehicleLocation(object):
    def __init__(self, latitude: float, longitude: float, altitude: float):
        self.latitude = latitude
        self.longitude = longitude
        self.altitude = altitude


class BLUvoConfig(object):

    def __init__(self, region: Regions, make: Makes, username: str, password: str, pin: str,
                 auto_login=False, vin=None, vehicle_id=None):
        self.vehicle_id = vehicle_id
        self.vin = vin
        self.auto_login = auto_login
        self.pin = pin
        self.password = password
        self.username = username
        self.make = make
        self.region = region

        self.base_url = ""


class BLUvoSession(object):

    def __init__(self, access_token, refresh_token, token_expires_in):
        now = datetime.now()
        timestamp = datetime.timestamp(now)

        self.token_expires_at = timestamp + token_expires_in
        self.refresh_token = refresh_token
        self.access_token = access_token
        self.pin_token = None

    def set_pin_token(self, token):
        self.pin_token = token


class VehicleInterface(object):

    @abstractmethod
    def __init__(self, bluvo_config: BLUvoConfig):
        self.bluvo_config = bluvo_config
        self.id = ""

    def __str__(self):
        return 'Vehicle: %s' % (self.id,)

    def __repr__(self):
        return 'Vehicle: %s' % (self.id,)

    def status(self, refresh=False) -> Either[VehicleStatus, str]:
        return Either[VehicleStatus, str].right("Not Implemented")

    def location(self, refresh=False) -> Either[VehicleLocation, str]:
        return Either[VehicleLocation, str].right("Not Implemented")


class BLUvoInterface(object):

    @abstractmethod
    def __init__(self, bluvo_config: BLUvoConfig):
        pass

    @abstractmethod
    def login(self):
        raise NotImplementedError

    @abstractmethod
    def verify_pin(self, pin: str):
        raise NotImplementedError

    @abstractmethod
    def get_vehicles(self) -> List[VehicleInterface]:
        raise NotImplementedError

    def get_vehicle(self, vehicle_id: str) -> VehicleInterface:
        return next(vehicle for vehicle in self.get_vehicles() if vehicle.id == vehicle_id)


class BLUvoInternalInterface(BLUvoInterface):

    @abstractmethod
    def request(self, cmd, data=None, headers=None) -> Either[dict, str]:
        raise NotImplementedError



from setuptools import setup, find_packages

setup(
    name='BLUvo',
    version='0.7',
    license='MIT',
    description='API Wrapper for the Bluelink/Uvo service',
    author='William Comartin',
    author_email='williamcomartin@gmail.com',
    url='https://github.com/wcomartin/bluvo',
    download_url='https://github.com/wcomartin/bluvo/archive/v_0.7.tar.gz',
    keywords=['Kia', 'Uvo', 'Api', 'bluelink'],
    packages=find_packages(),
    install_requires=[
        'requests',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



from typing import TypeVar, Mapping, Type, Callable, Any

TLeft = TypeVar('TLeft')
TRight = TypeVar('TRight')


class Either(Mapping[TLeft, TRight]):

    _left: Type[TLeft]
    _right: Type[TRight]
    _isLeft: bool

    def __getitem__(self, k):
        pass

    def __len__(self):
        pass

    def __iter__(self):
        pass

    @classmethod
    def left(cls, val: Type[TLeft]):
        either = Either[TLeft, TRight]()
        either._left = val
        either._isLeft = True
        return either

    @classmethod
    def right(cls, val: Type[TRight]):
        either = Either[TLeft, TRight]()
        either._right = val
        either._isLeft = False
        return either

    def match(self, left_func: Callable[[TLeft], Any], right_func: Callable[[TRight], Any]):
        return left_func(self._left) if self._isLeft else right_func(self._right)
